// --------------------------------------------------------------------------
//
//  Te2DSort.C -- TE/2 Dialing Directory Sort Program
//
//    CopyRight 1990-1992 by Oberon Software, Mankato, MN
//    All Rights Reserved
//
// --------------------------------------------------------------------------

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <conio.h>

#define INCL_DOS
#include <os2.h>

// --------------------------------------------------------------------------
// ---- Constants

  // ---- Parity designation codes for directory entries
#define NO_PARITY         0       // No Parity
#define ODD_PARITY        1       // Odd Parity
#define EVEN_PARITY       2       // Even Parity
#define MARK_PARITY       3       // Mark Parity
#define SPACE_PARITY      4       // Space Parity

  // ---- StopBit designation codes for directory entries
#define STOPBITS_1        0       // 1 Stopbit
#define STOPBITS_15       1       // 1.5 Stopbits
#define STOPBITS_2        2       // 2 Stopbits

  // ---- Protocol designation codes for directory entries
#define PROTOCOL_ASCII    0       // Ascii protocol (upload only)
#define PROTOCOL_XMODEM   1       // XModem (CRC)
#define PROTOCOL_XMODEM1K 2       // XModem-1K
#define PROTOCOL_YMODEM   3       // YModem
#define PROTOCOL_ZMODEM   4       // ZModem
#define PROTOCOL_YMODEMG  5       // YModem-G (not used before 1.00c)
#define PROTOCOL_CISQB    6       // CIS BPlus (for future use)

  // ---- Emulation designation codes for directory entries
#define EMULATE_TTY       0       // Teletype emulation
#define EMULATE_ANSI      1       // ANSI-BBS
#define EMULATE_XANSI     2       // ANSI-TE/2
#define EMULATE_VT100     3       // VT100

  // ---- Other important constants
#define ORIG_MAXDENTRY    50      // # of entries in pre-1.00c files
#define MAXDENTRY         200     // # of entries in files since 1.00c
#define TAGLEN            5       // Length of TAG field
#define NAMELEN           25      // Length of NAME field
#define NUMLEN            29      // Length of NUMBER field
#define SCRIPTLEN         63      // Length of SCRIPT field
#define ACCCODES          10      // Number of access code entries/file
#define ACCCLEN           NUMLEN  // Length of access code entry
#define RDF_BUFSIZ        80      // Length of a utility buffer


// --------------------------------------------------------------------------
// ---- Typedefs

  // ---- Directory entry structure
typedef struct _DDENTRY
{
  char     tag[TAGLEN+1];       // May be any asciiz data
  char     name[NAMELEN+1];     // May be any asciiz data
  char     number[NUMLEN+1];    // May be any asciiz data
  USHORT   baud;                // 110, 150, 300, 600, 1200, 2400, 4800, 9600,
                                //   19200, 38400, or 57600
  USHORT   parity;              // NO_PARITY, ODD_PARITY, EVEN_PARITY,
                                //   MARK_PARITY, or SPACE_PARITY
  USHORT   wordlen;             // 7 or 8
  USHORT   stopbits;            // STOPBITS_1, STOPBITS_15, or STOPBITS_2
  USHORT   duplex;              // TRUE == Half Duplex, FALSE == Full Duplex
  USHORT   protocol;            // PROTOCOL_ASCII, PROTOCOL_XMODEM,
                                //   PROTOCOL_XMODEM1K, PROTOCOL_YMODEM,
                                //   PROTOCOL_ZMODEM, PROTOCOL_YMODEMG, or
                                //   PROTOCOL_CISQB
  USHORT   emulate;             // EMULATE_TTY, EMULATE_ANSI, EMULATE_XANSI,
                                //   or EMULATE_VT100
  char     script[SCRIPTLEN+1]; // May be any asciiz data
  USHORT   connected;           // Count of times connected
  DATETIME lastconnect;         // Date/Time of last connect
  char     password[NAMELEN+1]; // May be any asciiz data
} DDENTRY;


// --------------------------------------------------------------------------
// ---- Local data

  // ---- Array of pointers to directory entries, will be allocated and
  //      filled by ReadDirectoryFile()
static DDENTRY *directory[MAXDENTRY];

  // ---- Array of string buffers to hold access codes associated with
  //      a dialing directory file.
static char accCodes[ACCCODES][ACCCLEN+1];

  // ---- Signiture strings found in the header record of a directory
  //      file from various versions of TE/2
static char *dirFileSig_v0    = "TE/2 Directory v1.0";
static char *dirFileSig_v100F = "TE/2 Directory v1.001";
static char *dirFileSig_v110D = "TE/2 Directory v1.002";
static char *dirFileSig       = "TE/2 Directory v1.10.E";

  // ---- An empty DATETIME structure used for default value of "lastconnect"
  //      in older directory files
static DATETIME nulDate   = { '\0', '\0', '\0', '\0', '\0', '\0', 0, 0, '\0' };

  // ---- Valid sort options
char options[]  = "0TNPD#";

// --------------------------------------------------------------------------

  // ---- Function prototypes

    // ---- Stuff directly from TE2DDIR.C
static int freadStrg(char *buf, FILE *fp);  // Read asciiz from file
static int fwriteStrg(char *buf, FILE *fp); // Write asciiz to file

static int ReadV0Directory(FILE *fp);       // Read a very old dir file
static int ReadV100FDirectory(FILE *fp);    // Read an old dir file
static int ReadCurrentDirectory(FILE *fp, BOOL f);  // Read a current dir file

int  ReadDirectoryFile(char *fname);        // Read any vsn dir file
int  WriteDirectoryFile(char *fname);       // Write a current vsn dir file

void EmptyDirectory(void);                  // free all allocated memory

    // ---- Added functions for this TE2DSORT.C
void Sort(int opt, BOOL fInvert);
int  main(int argc, char **argv);


// --------------------------------------------------------------------------
// ---- Functions


  // -----------------------------------------------------------------------
  // Read an ASCIIZ string from a file, expects 'buf' to be large enough
  // to hold any conceivable number of characters (i.e., there is no
  // error checking).

static int freadStrg(char *buf, FILE *fp)
{
  int  i = 0, err = FALSE;
  char c;

  do
  {
    if (fread(&c, sizeof(char), 1, fp) != 1)
    {
      err = TRUE;
      break;
    }
    buf[i++] = c;
  }
  while (c);

  return err;
}



  // -----------------------------------------------------------------------
  // Write an ASCIIZ string to a file

static int fwriteStrg(char *buf, FILE *fp)
{
  int  i = 0, err = FALSE;
  char c;

  do
  {
    c = buf[i++];
    if (fwrite(&c, sizeof(char), 1, fp) != 1)
    {
      err = TRUE;
      break;
    }
  }
  while (c);

  return err;
}


// --------------------------------------------------------------------------


  // -----------------------------------------------------------------------
  // General Notes about the Read*Directory() functions:
  //
  //   The "directory" array is an array of pointers to directory entry
  //   structures, it contains all NULL pointers by default.  Each
  //   function will allocate entries as needed as it encounters valid
  //   entries in the directory file.  Empty entries in the file consist
  //   of a single, zero byte.  When an empty entry is encountered, it's
  //   pointer in the array is left NULL.
  //
  //   These functions may be called more than once to reread the same
  //   or other directory files.  Space for entries will be reused,
  //   allocated, and/or freed as required.
  //
  //   Each function returns TRUE if the file was read with no errors
  //   detected, FALSE if any error occured.  If FALSE is returned, the
  //   directory may be partially read and available for examination.
  //
  //   After the caller is done with the "directory" array, the
  //   EmptyDirectory() function should be called to make sure that all
  //   allocated memory has been freed.
  // -----------------------------------------------------------------------


  // -----------------------------------------------------------------------
  // Read a "version 0" directory file into memory.  This is the file format
  // created by TE/2 Version 1.00b.  Expects the current position in the
  // file to be pointing past the header information at the first real entry.

static int ReadV0Directory(FILE *fp)
{
  int  i, j, err;
  char buf = 0;

  for (i = 0; i < ORIG_MAXDENTRY; i++)
  {
    if (fread(&buf, sizeof(char), 1, fp) == 1)
    {
      if (buf)  // If there is an entry at this position
      {
        // Either create an entry structure if this is a first read or
        //   clean up an existing one for reuse.
        if (!directory[i])
          directory[i] = calloc(1, sizeof(DDENTRY));
        else
          memset(directory[i], 0, sizeof(DDENTRY));

        // Read the data that's there
        err  =  freadStrg(directory[i]->tag, fp);
        err |=  freadStrg(directory[i]->name, fp);
        err |=  freadStrg(directory[i]->number, fp);
        err |= (fread(&(directory[i]->baud),     sizeof(USHORT), 1, fp) != 1);
        err |= (fread(&(directory[i]->parity),   sizeof(USHORT), 1, fp) != 1);
        err |= (fread(&(directory[i]->wordlen),  sizeof(USHORT), 1, fp) != 1);
        err |= (fread(&(directory[i]->stopbits), sizeof(USHORT), 1, fp) != 1);
        err |= (fread(&(directory[i]->duplex),   sizeof(USHORT), 1, fp) != 1);
        err |= (fread(&(directory[i]->protocol), sizeof(USHORT), 1, fp) != 1);
        err |=  freadStrg(directory[i]->script, fp);
        if (err) break;

        // Insert default values, this version of the file didn't contain
        //   these fields.
        directory[i]->emulate = EMULATE_XANSI;
        directory[i]->connected = 0;
        directory[i]->lastconnect = nulDate;
      }
      else if (directory[i])
      {
        // No entry at this position in this file.  If there was already
        // space allocated, free it.
        free(directory[i]);
        directory[i] = NULL;
      }
    }
    else
      break;
  }

  if (!err)
  {
    // This version of the file did not contain as many records as later
    // ones.  Make up the difference here with NULL entries.
    for (j = i; j < MAXDENTRY; j++)
      if (directory[j])
      {
        free(directory[j]);
        directory[j] = NULL;
      }

    // This version of the file didn't contain any Access Codes.  Fill in
    // with NULL entries.
    for (j = 0; j < ACCCODES; j++)
      memset(accCodes[j], 0, ACCCLEN+1);
  }

  // Return TRUE on a successful read, FALSE if an error occured
  return (i == ORIG_MAXDENTRY);
}




  // -----------------------------------------------------------------------
  // Read a "version 100F" directory file into memory.  This is the file
  // format created by TE/2 starting with Version 1.00c and continued
  // through version 1.00f. Expects the current position in the file to
  // be pointing past the header information at the first real entry.

static int ReadV100FDirectory(FILE *fp)
{
  int  i, j, err;
  char buf = 0;

  for (i = 0; i < MAXDENTRY; i++)
  {
    if (fread(&buf, sizeof(char), 1, fp) == 1)
    {
      if (buf)
      {
        // Either create an entry structure if this is a first read or
        //   clean up an existing one for reuse.
        if (!directory[i])
          directory[i] = calloc(1, sizeof(DDENTRY));
        else
          memset(directory[i], 0, sizeof(DDENTRY));

        // Read the data that's there
        err  =  freadStrg(directory[i]->tag, fp);
        err |=  freadStrg(directory[i]->name, fp);
        err |=  freadStrg(directory[i]->number, fp);
        err |= (fread(&(directory[i]->baud),     sizeof(USHORT), 1, fp) != 1);
        err |= (fread(&(directory[i]->parity),   sizeof(USHORT), 1, fp) != 1);
        err |= (fread(&(directory[i]->wordlen),  sizeof(USHORT), 1, fp) != 1);
        err |= (fread(&(directory[i]->stopbits), sizeof(USHORT), 1, fp) != 1);
        err |= (fread(&(directory[i]->duplex),   sizeof(USHORT), 1, fp) != 1);
        err |= (fread(&(directory[i]->protocol), sizeof(USHORT), 1, fp) != 1);
        err |= (fread(&(directory[i]->emulate),  sizeof(USHORT), 1, fp) != 1);
        err |=  freadStrg(directory[i]->script, fp);
        if (err) break;

        // Insert default values, this version of the file didn't contain
        //   these fields.
        directory[i]->connected = 0;
        directory[i]->lastconnect = nulDate;
      }
      else if (directory[i])
      {
        // No entry at this position in this file.  If there was already
        // space allocated, free it.
        free(directory[i]);
        directory[i] = NULL;
      }
    }
    else
      break;
  }

  if (!err)
  {
    // This version of the file didn't contain any Access Codes.  Fill in
    // with NULL entries.
    for (j = 0; j < ACCCODES; j++)
      memset(accCodes[j], 0, ACCCLEN+1);
  }

  // Return TRUE on a successful read, FALSE if an error occured
  return (i == MAXDENTRY);
}




  // -----------------------------------------------------------------------
  // Read a "current version" directory file into memory.  This is the
  // file format created by TE/2 starting with Version 1.10a. Expects
  // the current position in the file to be pointing past the header
  // information at the first real entry.
  // -----------------------------------------------------------------------
  // Note regarding the f110EFlag: The only change in the 1.10e file format
  // from the previous version is the including of the password field at the
  // end of the record.  We can use the same routine to ready both
  // versions.

static int ReadCurrentDirectory(FILE *fp, BOOL f110EFlag)
{
  int  i, j, err;
  char buf = 0;

  for (i = 0; i < MAXDENTRY; i++)
  {
    if (fread(&buf, sizeof(char), 1, fp) == 1)
    {
      if (buf)
      {
        // Either create an entry structure if this is a first read or
        //   clean up an existing one for reuse.
        if (!directory[i])
          directory[i] = calloc(1, sizeof(DDENTRY));
        else
          memset(directory[i], 0, sizeof(DDENTRY));

        // Read the data
        err  =  freadStrg(directory[i]->tag, fp);
        err |=  freadStrg(directory[i]->name, fp);
        err |=  freadStrg(directory[i]->number, fp);
        err |= (fread(&(directory[i]->baud),        sizeof(USHORT),   1, fp) != 1);
        err |= (fread(&(directory[i]->parity),      sizeof(USHORT),   1, fp) != 1);
        err |= (fread(&(directory[i]->wordlen),     sizeof(USHORT),   1, fp) != 1);
        err |= (fread(&(directory[i]->stopbits),    sizeof(USHORT),   1, fp) != 1);
        err |= (fread(&(directory[i]->duplex),      sizeof(USHORT),   1, fp) != 1);
        err |= (fread(&(directory[i]->protocol),    sizeof(USHORT),   1, fp) != 1);
        err |= (fread(&(directory[i]->emulate),     sizeof(USHORT),   1, fp) != 1);
        err |=  freadStrg(directory[i]->script, fp);
        err |= (fread(&(directory[i]->connected),   sizeof(USHORT),   1, fp) != 1);
        err |= (fread(&(directory[i]->lastconnect), sizeof(DATETIME), 1, fp) != 1);
        if (f110EFlag) err |=  freadStrg(directory[i]->password, fp);
        if (err) break;
      }
      else if (directory[i])
      {
        // No entry at this position in this file.  If there was already
        // space allocated, free it.
        free(directory[i]);
        directory[i] = NULL;
      }
    }
    else
      break;
  }

  if (!err && !feof(fp))
  {
    // Read the Access Codes associated with this directory
    for (j = 0; j < ACCCODES && !err; j++)
      err |= freadStrg(accCodes[j], fp);
  }

  // Return TRUE on a successful read, FALSE if an error occured
  return ((i == MAXDENTRY) && !err);
}




  // -----------------------------------------------------------------------
  // Read any format directory file into memory.

int ReadDirectoryFile(char *fname)
{
  FILE *fp;
  int  i, rVal = FALSE;
  char buf[RDF_BUFSIZ];

  if (fp = fopen(fname, "rb"))
  {
    // "dirFileSig", the signiture on a current version directory file is
    // the longest signiture in any version.  We'll read enough of the
    // header to get the whole string; we may need to seek back in the
    // file if it's an older version.

    i = strlen(dirFileSig);
    if ((int)fread(buf, sizeof(char), i, fp) == i)
    {
      if (strncmp(buf, dirFileSig, i) == 0)
      {
        // Current style directory, Version 1.10.e on
        rVal = ReadCurrentDirectory(fp, TRUE);
      }
      else if (strncmp(buf, dirFileSig_v110D, strlen(dirFileSig_v110D)) == 0)
      {
        // Version 1.10.a through 1.10.d style directory
        rVal = ReadCurrentDirectory(fp, FALSE);
      }
      else if (strncmp(buf, dirFileSig_v100F, strlen(dirFileSig_v100F)) == 0)
      {
        // Version 1.00.c through 1.00.f style directory
        fseek(fp, (long)(strlen(dirFileSig_v100F)), SEEK_SET);
        rVal = ReadV100FDirectory(fp);
      }
      else if (strncmp(buf, dirFileSig_v0, strlen(dirFileSig_v0)) == 0)
      {
        // Earliest style, first alpha
        fseek(fp, (long)(strlen(dirFileSig_v0)), SEEK_SET);
        rVal = ReadV0Directory(fp);
      }
    }
    fclose(fp);
  }

  return rVal;
}



  // -----------------------------------------------------------------------
  // Write a directory file of the current format.

int WriteDirectoryFile(char *fname)
{
  FILE *fp;
  int  i, j, err, rVal = FALSE;
  char c;

  if (fp = fopen(fname, "wb"))
  {
    i = strlen(dirFileSig);
    if ((int)fwrite(dirFileSig, sizeof(char), i, fp) == i)
    {
      for (i = 0; i < MAXDENTRY; i++)
      {
        if (directory[i])
        {
          c = 1;
          if (fwrite(&c, sizeof(char), 1, fp) == 1)
          {
            err  =  fwriteStrg(directory[i]->tag, fp);
            err |=  fwriteStrg(directory[i]->name, fp);
            err |=  fwriteStrg(directory[i]->number, fp);
            err |= (fwrite(&(directory[i]->baud),        sizeof(USHORT),   1, fp) != 1);
            err |= (fwrite(&(directory[i]->parity),      sizeof(USHORT),   1, fp) != 1);
            err |= (fwrite(&(directory[i]->wordlen),     sizeof(USHORT),   1, fp) != 1);
            err |= (fwrite(&(directory[i]->stopbits),    sizeof(USHORT),   1, fp) != 1);
            err |= (fwrite(&(directory[i]->duplex),      sizeof(USHORT),   1, fp) != 1);
            err |= (fwrite(&(directory[i]->protocol),    sizeof(USHORT),   1, fp) != 1);
            err |= (fwrite(&(directory[i]->emulate),     sizeof(USHORT),   1, fp) != 1);
            err |=  fwriteStrg(directory[i]->script, fp);
            err |= (fwrite(&(directory[i]->connected),   sizeof(USHORT),   1, fp) != 1);
            err |= (fwrite(&(directory[i]->lastconnect), sizeof(DATETIME), 1, fp) != 1);
            err |=  fwriteStrg(directory[i]->password, fp);
            if (err) break;
          }
          else
            break;
        }
        else
        {
          c = 0;
          if (fwrite(&c, sizeof(char), 1, fp) != 1)
            break;
        }
      }
      rVal = (i == MAXDENTRY);

      if (!err && !feof(fp))
      {
        for (j = 0; j < ACCCODES; j++)
          err |= fwriteStrg(accCodes[j], fp);
      }

    }
    fclose(fp);
  }

  return rVal;
}



  // -----------------------------------------------------------------------
  // Free all allocated memory associated with the directory


void EmptyDirectory()
{
  int i;

  for (i = 0; i < MAXDENTRY; i++)
  {
    if (directory[i])
    {
      free(directory[i]);
      directory[i] = NULL;
    }
  }
}



// --------------------------------------------------------------------------


void Sort(int opt, BOOL fInvert)
{
  // Bubble Sort.  It should be quick enough on 200 items and it has the
  // added feature of being a stable sort so that subsequent resorts will
  // behave as advertised.

  BOOL    fDone = FALSE;
  BOOL    fSwap;
  DDENTRY *p;
  char    buf1[16], buf2[16];
  USHORT  i;

  while (!fDone)
  {
    fDone = TRUE;
    for (i = 1; i < MAXDENTRY; i++)
    {
      // NULL entries are 'heavier' than non-NULL
      fSwap = (directory[i] && !directory[i-1]);

      // If this entry is NULL it cannot be 'lighter' than the previous,
      // at best it's the same so we won't force a swap.  Likewise, if
      // we've already decided to swap because the previous entry was NULL,
      // there's no reason to check further.

      if (directory[i] && !fSwap)
      {
        switch (opt)
        {
          case 'T':
            if (fInvert)
              fSwap = (strcmp(directory[i-1]->tag, directory[i]->tag) < 0);
            else
              fSwap = (strcmp(directory[i-1]->tag, directory[i]->tag) > 0);
            break;

          case 'N':
            if (fInvert)
              fSwap = (strcmp(directory[i-1]->name, directory[i]->name) < 0);
            else
              fSwap = (strcmp(directory[i-1]->name, directory[i]->name) > 0);
            break;

          case 'P':
            if (fInvert)
              fSwap = (strcmp(directory[i-1]->number, directory[i]->number) < 0);
            else
              fSwap = (strcmp(directory[i-1]->number, directory[i]->number) > 0);
            break;

          case 'D':
            sprintf(buf1, "%04d%02d%02d%02d%02d%02d",
                          directory[i-1]->lastconnect.year,
                          directory[i-1]->lastconnect.month,
                          directory[i-1]->lastconnect.day,
                          directory[i-1]->lastconnect.hours,
                          directory[i-1]->lastconnect.minutes,
                          directory[i-1]->lastconnect.seconds);
            sprintf(buf2, "%04d%02d%02d%02d%02d%02d",
                          directory[i]->lastconnect.year,
                          directory[i]->lastconnect.month,
                          directory[i]->lastconnect.day,
                          directory[i]->lastconnect.hours,
                          directory[i]->lastconnect.minutes,
                          directory[i]->lastconnect.seconds);
            if (fInvert)
              fSwap = (strcmp(buf1, buf2) < 0);
            else
              fSwap = (strcmp(buf1, buf2) > 0);
            break;

          case '#':
            if (fInvert)
              fSwap = (directory[i]->connected > directory[i-1]->connected);
            else
              fSwap = (directory[i]->connected < directory[i-1]->connected);
            break;
        }
      }
      if (fSwap)
      {
        p              = directory[i];
        directory[i]   = directory[i-1];
        directory[i-1] = p;
        fDone = FALSE;
      }
    }
  }
}




int main(int argc, char **argv)
{
  BOOL fInvert, fParmError = FALSE;
  FILE *fp;
  char *cp, c;

  printf("TE2DSort - CopyRight 1990-1992, Oberon Software\n\n");

  if (argc < 4)
    fParmError = TRUE;
  else
  {
    // Make sure that all option characters are valid
    strupr(argv[3]);
    for (cp = argv[3]; *cp; cp++)
    {
      if ((strchr(options, *cp) == NULL) && (*cp != '-'))
      {
        fParmError = TRUE;
        break;
      }
    }

    // Make sure that the last option character ISN'T a hyphen
    if (!fParmError && (argv[3][strlen(argv[3])-1] == '-'))
      fParmError = TRUE;
  }

  if (fParmError)
  {
    fprintf(stderr, "usage: te2dsort dirfilename outfilename sortoption\n");
    fprintf(stderr, " where sortoption is one or more of:\n");
    fprintf(stderr, "       0 - No sort but close up NULL entries\n");
    fprintf(stderr, "       T - Sort by Tag\n");
    fprintf(stderr, "       N - Sort by Name\n");
    fprintf(stderr, "       P - Sort by Phonenumber\n");
    fprintf(stderr, "       D - Sort by Date of last connection\n");
    fprintf(stderr, "       # - Sort by # of times connected\n");
    fprintf(stderr, " any option character may be preceded with an hyphen to\n");
    fprintf(stderr, "   reverse the sort order on that option only.\n\n");
    fprintf(stderr, " examples:\n");
    fprintf(stderr, "   te2dsort te2.dir te2dir.new T\n");
    fprintf(stderr, "     will sort te2.dir by Tag and write output to te2dir.new\n");
    fprintf(stderr, "   te2dsort te2.dir te2dir.new -D\n");
    fprintf(stderr, "     will sort te2.dir by date, newest to oldest\n");
    fprintf(stderr, "   te2dsort te2.dir te2dir.new PN\n");
    fprintf(stderr, "     will sort first by Phonenumber and then by Name before\n");
    fprintf(stderr, "     writing the output to te2dir.new.  The result will be\n");
    fprintf(stderr, "     a file sorted primarily by Name and subsorted by Phonenumber.\n\n");
    exit(1);
  }

  if (!ReadDirectoryFile(argv[1]))
  {
    fprintf(stderr, "Error reading file \"%s\"\n", argv[1]);
    exit(1);
  }

  fInvert = FALSE;
  for (cp = argv[3]; *cp; cp++)
  {
    if (*cp == '-')
      fInvert = TRUE;
    else
    {
      printf("Sorting using option %c%s ...", *cp, fInvert ? " (inverted)" : "");
      Sort(*cp, fInvert);
      fInvert = FALSE;
      printf(" Done.\n");
    }
  }

  if (fp = fopen(argv[2], "r"))
  {
    fclose(fp);
    fprintf(stderr, "File: \"%s\" already exists! Overwrite? [y/N] ", argv[2]);
    while (TRUE)
    {
      c = (char)getch();
      if (cp = strchr("yYnN\x1b\n", c))
      {
        if ((*cp == 'y') || (*cp == 'Y'))
        {
          fprintf(stderr, "Yes\n");
          break;
        }
        else
        {
          fprintf(stderr, "No\nAborted\n");
          exit(1);
        }
      }
    }
  }

  if (!WriteDirectoryFile(argv[2]))
  {
    fprintf(stderr, "Error writing file \"%s\"\n", argv[2]);
    exit(1);
  }

  return 0;
}

